@@ -4,6 +4,10 @@ namespace Tests\Prophecy\Doubler\Generator;
 
 use PHPUnit\Framework\TestCase;
 use Prophecy\Doubler\Generator\ClassMirror;
+use Prophecy\Doubler\Generator\Node\ArgumentTypeNode;
+use Prophecy\Doubler\Generator\Node\ReturnTypeNode;
+use Prophecy\Exception\Doubler\ClassMirrorException;
+use Prophecy\Exception\InvalidArgumentException;
 
 class ClassMirrorTest extends TestCase
 {
@@ -75,7 +79,7 @@ class ClassMirrorTest extends TestCase
         $this->assertCount(1, $argNodes);
 
         $this->assertEquals('arg', $argNodes[0]->getName());
-        $this->assertNull($argNodes[0]->getTypeHint());
+        $this->assertEquals(new ArgumentTypeNode(), $argNodes[0]->getTypeNode());
         $this->assertFalse($argNodes[0]->isOptional());
         $this->assertNull($argNodes[0]->getDefault());
         $this->assertFalse($argNodes[0]->isPassedByReference());
@@ -97,19 +101,20 @@ class ClassMirrorTest extends TestCase
 
         $this->assertCount(3, $argNodes);
 
+
         $this->assertEquals('arg_1', $argNodes[0]->getName());
-        $this->assertEquals('array', $argNodes[0]->getTypeHint());
-        $this->assertTrue($argNodes[0]->isOptional());
-        $this->assertEquals(array(), $argNodes[0]->getDefault());
-        $this->assertFalse($argNodes[0]->isPassedByReference());
-        $this->assertFalse($argNodes[0]->isVariadic());
+        $this->assertEquals(new ArgumentTypeNode('ArrayAccess'), $argNodes[0]->getTypeNode());
+        $this->assertFalse($argNodes[0]->isOptional());
 
         $this->assertEquals('arg_2', $argNodes[1]->getName());
-        $this->assertEquals('ArrayAccess', $argNodes[1]->getTypeHint());
-        $this->assertFalse($argNodes[1]->isOptional());
+        $this->assertEquals(new ArgumentTypeNode('array'), $argNodes[1]->getTypeNode());
+        $this->assertTrue($argNodes[1]->isOptional());
+        $this->assertEquals(array(), $argNodes[1]->getDefault());
+        $this->assertFalse($argNodes[1]->isPassedByReference());
+        $this->assertFalse($argNodes[1]->isVariadic());
 
         $this->assertEquals('arg_3', $argNodes[2]->getName());
-        $this->assertEquals('ArrayAccess', $argNodes[2]->getTypeHint());
+        $this->assertEquals(new ArgumentTypeNode('ArrayAccess', 'null'), $argNodes[2]->getTypeNode());
         $this->assertTrue($argNodes[2]->isOptional());
         $this->assertNull($argNodes[2]->getDefault());
         $this->assertFalse($argNodes[2]->isPassedByReference());
@@ -133,13 +138,13 @@ class ClassMirrorTest extends TestCase
         $this->assertCount(2, $argNodes);
 
         $this->assertEquals('arg_1', $argNodes[0]->getName());
-        $this->assertEquals('callable', $argNodes[0]->getTypeHint());
+        $this->assertEquals(new ArgumentTypeNode('callable'), $argNodes[0]->getTypeNode());
         $this->assertFalse($argNodes[0]->isOptional());
         $this->assertFalse($argNodes[0]->isPassedByReference());
         $this->assertFalse($argNodes[0]->isVariadic());
 
         $this->assertEquals('arg_2', $argNodes[1]->getName());
-        $this->assertEquals('callable', $argNodes[1]->getTypeHint());
+        $this->assertEquals(new ArgumentTypeNode('callable', 'null'), $argNodes[1]->getTypeNode());
         $this->assertTrue($argNodes[1]->isOptional());
         $this->assertNull($argNodes[1]->getDefault());
         $this->assertFalse($argNodes[1]->isPassedByReference());
@@ -163,7 +168,7 @@ class ClassMirrorTest extends TestCase
         $this->assertCount(1, $argNodes);
 
         $this->assertEquals('args', $argNodes[0]->getName());
-        $this->assertNull($argNodes[0]->getTypeHint());
+        $this->assertEquals(new ArgumentTypeNode(), $argNodes[0]->getTypeNode());
         $this->assertFalse($argNodes[0]->isOptional());
         $this->assertFalse($argNodes[0]->isPassedByReference());
         $this->assertTrue($argNodes[0]->isVariadic());
@@ -190,7 +195,7 @@ class ClassMirrorTest extends TestCase
         $this->assertCount(1, $argNodes);
 
         $this->assertEquals('args', $argNodes[0]->getName());
-        $this->assertEquals('array', $argNodes[0]->getTypeHint());
+        $this->assertEquals(new ArgumentTypeNode('array'), $argNodes[0]->getTypeNode());
         $this->assertFalse($argNodes[0]->isOptional());
         $this->assertFalse($argNodes[0]->isPassedByReference());
         $this->assertTrue($argNodes[0]->isVariadic());
@@ -219,7 +224,6 @@ class ClassMirrorTest extends TestCase
 
     /**
      * @test
-     * @expectedException Prophecy\Exception\Doubler\ClassMirrorException
      */
     public function it_throws_an_exception_if_class_is_final()
     {
@@ -227,6 +231,8 @@ class ClassMirrorTest extends TestCase
 
         $mirror = new ClassMirror();
 
+        $this->expectException(ClassMirrorException::class);
+
         $mirror->reflect($class, array());
     }
 
@@ -261,7 +267,6 @@ class ClassMirrorTest extends TestCase
 
     /**
      * @test
-     * @expectedException Prophecy\Exception\InvalidArgumentException
      */
     public function it_throws_an_exception_if_interface_provided_instead_of_class()
     {
@@ -269,6 +274,8 @@ class ClassMirrorTest extends TestCase
 
         $mirror = new ClassMirror();
 
+        $this->expectException(InvalidArgumentException::class);
+
         $mirror->reflect($class, array());
     }
 
@@ -345,14 +352,13 @@ class ClassMirrorTest extends TestCase
         $this->assertTrue($classNode->hasMethod('getSelf'));
         $this->assertTrue($classNode->hasMethod('getParent'));
 
-        $this->assertEquals('string', $classNode->getMethod('getName')->getReturnType());
-        $this->assertEquals('\Fixtures\Prophecy\WithReturnTypehints', $classNode->getMethod('getSelf')->getReturnType());
-        $this->assertEquals('\Fixtures\Prophecy\EmptyClass', $classNode->getMethod('getParent')->getReturnType());
+        $this->assertEquals(new ReturnTypeNode('string'), $classNode->getMethod('getName')->getReturnTypeNode());
+        $this->assertEquals(new ReturnTypeNode('\Fixtures\Prophecy\WithReturnTypehints'), $classNode->getMethod('getSelf')->getReturnTypeNode());
+        $this->assertEquals(new ReturnTypeNode('\Fixtures\Prophecy\EmptyClass'), $classNode->getMethod('getParent')->getReturnTypeNode());
     }
 
     /**
      * @test
-     * @expectedException InvalidArgumentException
      */
     public function it_throws_an_exception_if_class_provided_in_interfaces_list()
     {
@@ -360,17 +366,20 @@ class ClassMirrorTest extends TestCase
 
         $mirror = new ClassMirror();
 
+        $this->expectException(\InvalidArgumentException::class);
+
         $mirror->reflect(null, array($class));
     }
 
     /**
      * @test
-     * @expectedException InvalidArgumentException
      */
     public function it_throws_an_exception_if_not_reflection_provided_as_interface()
     {
         $mirror = new ClassMirror();
 
+        $this->expectException(\InvalidArgumentException::class);
+
         $mirror->reflect(null, array(null));
     }
 
@@ -379,15 +388,19 @@ class ClassMirrorTest extends TestCase
      */
     public function it_doesnt_use_scalar_typehints()
     {
+        if (PHP_VERSION_ID >= 80000) {
+            $this->markTestSkipped('Internal types have scalar hints in PHP 8');
+        }
+
         $mirror = new ClassMirror();
 
         $classNode = $mirror->reflect(new \ReflectionClass('ReflectionMethod'), array());
         $method = $classNode->getMethod('export');
         $arguments = $method->getArguments();
 
-        $this->assertNull($arguments[0]->getTypeHint());
-        $this->assertNull($arguments[1]->getTypeHint());
-        $this->assertNull($arguments[2]->getTypeHint());
+        $this->assertEquals(new ArgumentTypeNode(), $arguments[0]->getTypeNode());
+        $this->assertEquals(new ArgumentTypeNode(), $arguments[1]->getTypeNode());
+        $this->assertEquals(new ArgumentTypeNode(), $arguments[2]->getTypeNode());
     }
 
     /**
@@ -400,7 +413,7 @@ class ClassMirrorTest extends TestCase
         $classNode = $mirror->reflect(new \ReflectionClass('Fixtures\Prophecy\OptionalDepsClass'), array());
         $method = $classNode->getMethod('iHaveAStrangeTypeHintedArg');
         $arguments = $method->getArguments();
-        $this->assertEquals('I\Simply\Am\Nonexistent', $arguments[0]->getTypeHint());
+        $this->assertEquals(new ArgumentTypeNode('I\Simply\Am\Nonexistent'), $arguments[0]->getTypeNode());
     }
 
     /**
@@ -414,8 +427,7 @@ class ClassMirrorTest extends TestCase
         $classNode = $mirror->reflect(new \ReflectionClass('Fixtures\Prophecy\NullableArrayParameter'), array());
         $method = $classNode->getMethod('iHaveNullableArrayParameterWithNotNullDefaultValue');
         $arguments = $method->getArguments();
-        $this->assertSame('array', $arguments[0]->getTypeHint());
-        $this->assertTrue($arguments[0]->isNullable());
+        $this->assertEquals(new ArgumentTypeNode('array', 'null'), $arguments[0]->getTypeNode());
     }
 
     /**
@@ -428,7 +440,7 @@ class ClassMirrorTest extends TestCase
         $classNode = $mirror->reflect(new \ReflectionClass('Fixtures\Prophecy\OptionalDepsClass'), array());
         $method = $classNode->getMethod('iHaveAnEvenStrangerTypeHintedArg');
         $arguments = $method->getArguments();
-        $this->assertEquals('I\Simply\Am\Not', $arguments[0]->getTypeHint());
+        $this->assertEquals(new ArgumentTypeNode('I\Simply\Am\Not'), $arguments[0]->getTypeNode());
     }
 
     /**
@@ -450,6 +462,16 @@ class ClassMirrorTest extends TestCase
         $this->assertCount(2, $method->getArguments());
     }
 
+    /**
+     * @test
+     * @doesNotPerformAssertions
+     */
+    function it_doesnt_fail_to_mock_self_referencing_interface()
+    {
+        $class = $this->prophesize('Fixtures\Prophecy\SelfReferencing');
+        $class->reveal();
+    }
+
     /**
      * @test
      */
@@ -472,10 +494,7 @@ class ClassMirrorTest extends TestCase
         $method->isProtected()->willReturn(false);
         $method->isStatic()->willReturn(false);
         $method->returnsReference()->willReturn(false);
-
-        if (version_compare(PHP_VERSION, '7.0', '>=')) {
-            $method->hasReturnType()->willReturn(false);
-        }
+        $method->hasReturnType()->willReturn(false);
 
         $parameter->getName()->willReturn('...');
         $parameter->isDefaultValueAvailable()->willReturn(true);
@@ -483,9 +502,10 @@ class ClassMirrorTest extends TestCase
         $parameter->isPassedByReference()->willReturn(false);
         $parameter->allowsNull()->willReturn(true);
         $parameter->getClass()->willReturn($class);
-        if (version_compare(PHP_VERSION, '5.6', '>=')) {
-            $parameter->isVariadic()->willReturn(false);
-        }
+        $parameter->getType()->willReturn(null);
+        $parameter->hasType()->willReturn(false);
+        $parameter->isVariadic()->willReturn(false);
+        $parameter->getDeclaringClass()->willReturn($class);
 
         $mirror = new ClassMirror();
 
@@ -499,6 +519,55 @@ class ClassMirrorTest extends TestCase
         $this->assertEquals('__dot_dot_dot__', $argumentNode->getName());
     }
 
+    /** @test */
+    public function it_can_double_a_class_with_union_return_types()
+    {
+        if (PHP_VERSION_ID < 80000) {
+            $this->markTestSkipped('Union types are not supported in this PHP version');
+        }
+
+        $classNode = (new ClassMirror())->reflect(new \ReflectionClass('Fixtures\Prophecy\UnionReturnTypes'), []);
+        $methodNode = $classNode->getMethods()['doSomething'];
+
+        $this->assertSame(['\stdClass', 'bool'], $methodNode->getReturnTypeNode()->getTypes());
+    }
+
+    /** @test */
+    public function it_can_double_a_class_with_union_argument_types()
+    {
+        if (PHP_VERSION_ID < 80000) {
+            $this->markTestSkipped('Union types are not supported in this PHP version');
+        }
+
+        $classNode = (new ClassMirror())->reflect(new \ReflectionClass('Fixtures\Prophecy\UnionArgumentTypes'), []);
+        $methodNode = $classNode->getMethods()['doSomething'];
+
+        $this->assertEquals(new ArgumentTypeNode('bool', '\\stdClass'), $methodNode->getArguments()[0]->getTypeNode());
+    }
+
+    /** @test */
+    public function it_can_double_a_class_with_mixed_types()
+    {
+        if (PHP_VERSION_ID < 80000) {
+            $this->markTestSkipped('Mixed type is not supported in this PHP version');
+        }
+
+        $classNode = (new ClassMirror())->reflect(new \ReflectionClass('Fixtures\Prophecy\MixedTypes'), []);
+        $methodNode = $classNode->getMethods()['doSomething'];
+
+        $this->assertEquals(new ArgumentTypeNode('mixed'), $methodNode->getArguments()[0]->getTypeNode());
+        $this->assertEquals(new ReturnTypeNode('mixed'), $methodNode->getReturnTypeNode());
+    }
+
+    /** @test */
+    public function it_can_double_inherited_self_return_type()
+    {
+        $classNode = (new ClassMirror())->reflect(new \ReflectionClass('Fixtures\Prophecy\ClassExtendAbstractWithMethodWithReturnType'), []);
+        $methodNode = $classNode->getMethods()['returnSelf'];
+
+        $this->assertEquals(new ReturnTypeNode('Fixtures\Prophecy\AbstractBaseClassWithMethodWithReturnType'), $methodNode->getReturnTypeNode());
+    }
+
     /**
      * @test
      */
